"""A runner for Markdown-based tests for ty"""
# /// script
# requires-python = ">=3.11"
# dependencies = [
#     "rich",
#     "watchfiles>=1.1.0",
# ]
# ///

from __future__ import annotations

import argparse
import json
import os
import subprocess
import sys
import threading
from pathlib import Path
from typing import Final, Literal, assert_never

from rich.console import Console
from watchfiles import Change, watch

CRATE_NAME: Final = "ty_python_semantic"
CRATE_ROOT: Final = Path(__file__).resolve().parent
TY_VENDORED: Final = CRATE_ROOT.parent / "ty_vendored"
DIRS_TO_WATCH: Final = (
    CRATE_ROOT,
    TY_VENDORED,
    CRATE_ROOT.parent / "ty_test/src",
)
MDTEST_DIR: Final = CRATE_ROOT / "resources" / "mdtest"
SNAPSHOTS_DIR: Final = MDTEST_DIR / "snapshots"
MDTEST_README: Final = CRATE_ROOT / "resources" / "README.md"


class MDTestRunner:
    mdtest_executable: Path | None
    console: Console
    filters: list[str]
    enable_external: bool
    upgrade_lockfiles: bool

    def __init__(
        self,
        filters: list[str] | None,
        enable_external: bool,
        upgrade_lockfiles: bool,
    ) -> None:
        self.mdtest_executable = None
        self.console = Console()
        self.filters = [
            f.removesuffix(".md").replace("/", "_").replace("-", "_")
            for f in (filters or [])
        ]
        self.enable_external = enable_external
        self.upgrade_lockfiles = upgrade_lockfiles

    def _run_cargo_test(self, *, message_format: Literal["human", "json"]) -> str:
        return subprocess.check_output(
            [
                "cargo",
                "test",
                "--package",
                CRATE_NAME,
                "--no-run",
                "--color=always",
                "--test=mdtest",
                "--message-format",
                message_format,
            ],
            cwd=CRATE_ROOT,
            env=dict(os.environ, CLI_COLOR="1"),
            stderr=subprocess.STDOUT,
            text=True,
        )

    def _recompile_tests(
        self, status_message: str, *, message_on_success: bool = True
    ) -> bool:
        with self.console.status(status_message):
            # Run it with 'human' format in case there are errors:
            try:
                self._run_cargo_test(message_format="human")
            except subprocess.CalledProcessError as e:
                print(e.output)
                return False

            # Run it again with 'json' format to find the mdtest executable:
            try:
                json_output = self._run_cargo_test(message_format="json")
            except subprocess.CalledProcessError as _:
                # `cargo test` can still fail if something changed in between the two runs.
                # Here we don't have a human-readable output, so just show a generic message:
                self.console.print("[red]Error[/red]: Failed to compile tests")
                return False

            if json_output:
                self._get_executable_path_from_json(json_output)

        if message_on_success:
            self.console.print("[dim]Tests compiled successfully[/dim]")
        return True

    def _get_executable_path_from_json(self, json_output: str) -> None:
        for json_line in json_output.splitlines():
            try:
                data = json.loads(json_line)
            except json.JSONDecodeError:
                continue
            if data.get("target", {}).get("name") == "mdtest":
                self.mdtest_executable = Path(data["executable"])
                break
        else:
            raise RuntimeError(
                "Could not find mdtest executable after successful compilation"
            )

    def _run_mdtest(
        self, arguments: list[str] | None = None, *, capture_output: bool = False
    ) -> subprocess.CompletedProcess:
        assert self.mdtest_executable is not None

        arguments = arguments or []
        return subprocess.run(
            [self.mdtest_executable, *arguments],
            cwd=CRATE_ROOT,
            env=dict(
                os.environ,
                CLICOLOR_FORCE="1",
                INSTA_FORCE_PASS="1",
                INSTA_OUTPUT="none",
                MDTEST_EXTERNAL="1" if self.enable_external else "0",
                MDTEST_UPGRADE_LOCKFILES="1" if self.upgrade_lockfiles else "0",
            ),
            capture_output=capture_output,
            text=True,
            check=False,
        )

    @staticmethod
    def _md_file_for_snapshot(snapshot_path: Path) -> Path | None:
        """Given a deleted .snap.new path, find the source .md file it belongs to.

        Snapshot filenames follow the pattern:
            <md_filename>_-_<Section_title>_-_…_(<hash>).snap.new
        where <md_filename> may be truncated to 20 characters with a trailing
        Unicode ellipsis (U+2026).
        """
        snapshot_name = snapshot_path.name.removesuffix(".new").removesuffix(".snap")

        md_filename = snapshot_name.split("_-_", 1)[0]

        is_truncated = md_filename.endswith("\u2026")
        if is_truncated:
            md_filename = md_filename[:-1]

        for md_file in MDTEST_DIR.rglob("*.md"):
            if is_truncated:
                if md_file.name.startswith(md_filename):
                    return md_file.relative_to(MDTEST_DIR)
            elif md_file.name == md_filename:
                return md_file.relative_to(MDTEST_DIR)
        return None

    def _run_mdtests_for_file(self, markdown_file: Path) -> None:
        test_name = f"mdtest::{markdown_file}"

        output = self._run_mdtest(["--exact", test_name], capture_output=True)

        if output.returncode == 0:
            if "running 0 tests\n" in output.stdout:
                self.console.log(
                    f"[yellow]Warning[/yellow]: No tests were executed with filter '{test_name}'"
                )
            else:
                self.console.print(
                    f"Test for [bold green]{markdown_file}[/bold green] succeeded"
                )
        else:
            self.console.print()
            self.console.rule(
                f"Test for [bold red]{markdown_file}[/bold red] failed",
                style="gray",
            )
            self._print_trimmed_cargo_test_output(
                output.stdout + output.stderr, test_name
            )

    def _print_trimmed_cargo_test_output(self, output: str, test_name: str) -> None:
        # Skip 'cargo test' boilerplate at the beginning:
        lines = output.splitlines()
        start_index = 0
        for i, line in enumerate(lines):
            if f"{test_name} stdout" in line:
                start_index = i
                break

        for line in lines[start_index + 1 :]:
            if "MDTEST_TEST_FILTER" in line:
                continue
            if line.strip() == "-" * 50:
                # Skip 'cargo test' boilerplate at the end
                break

            print(line)

    def watch(self):
        def keyboard_input() -> None:
            for _ in sys.stdin:
                # This is silly, but there is no other way to inject events into
                # the main `watch` loop. We use changes to the `README.md` file
                # as a trigger to re-run all mdtests:
                MDTEST_README.touch()

        input_thread = threading.Thread(target=keyboard_input, daemon=True)
        input_thread.start()

        self._recompile_tests("Compiling tests...", message_on_success=False)
        self._run_mdtest(self.filters)
        self.console.print("[dim]Ready to watch for changes...[/dim]")

        for changes in watch(*DIRS_TO_WATCH):
            new_md_files: set[Path] = set()
            changed_md_files: set[Path] = set()
            rejected_snapshot_md_files: set[Path] = set()
            rust_code_has_changed = False
            vendored_typeshed_has_changed = False

            for change, path_str in changes:
                path = Path(path_str)

                # See above: `README.md` changes trigger a full re-run of all tests
                if path == MDTEST_README:
                    self._run_mdtest(self.filters)
                    continue

                # When a pending snapshot (.snap.new) is rejected in a separate
                # process (e.g. `cargo insta review`), the file is deleted.
                # A common reason for rejecting a snapshot is that it was stale
                # (produced by an earlier test run and now outdated). Re-run the
                # relevant tests so the snapshot is regenerated from the current
                # state of the code.
                if (
                    change == Change.deleted
                    and path.name.endswith(".snap.new")
                    and path.is_relative_to(SNAPSHOTS_DIR)
                ):
                    md_file = self._md_file_for_snapshot(path)
                    if md_file is not None:
                        rejected_snapshot_md_files.add(md_file)
                    continue

                match path.suffix:
                    case ".rs":
                        rust_code_has_changed = True
                    case ".pyi" if path.is_relative_to(TY_VENDORED):
                        vendored_typeshed_has_changed = True
                    case ".md":
                        pass
                    case _:
                        continue

                try:
                    relative_path = Path(path).relative_to(MDTEST_DIR)
                except ValueError:
                    continue

                match change:
                    case Change.added:
                        # When saving a file, some editors (looking at you, Vim) might first
                        # save the file with a temporary name (e.g. `file.md~`) and then rename
                        # it to the final name. This creates a `deleted` and `added` change.
                        # We treat those files as `changed` here.
                        if (Change.deleted, path_str) in changes:
                            changed_md_files.add(relative_path)
                        else:
                            new_md_files.add(relative_path)
                    case Change.modified:
                        changed_md_files.add(relative_path)
                    case Change.deleted:
                        # No need to do anything when a Markdown test is deleted
                        pass
                    case _ as unreachable:
                        assert_never(unreachable)

            if rust_code_has_changed:
                if self._recompile_tests("Rust code has changed, recompiling tests..."):
                    self._run_mdtest(self.filters)
            elif vendored_typeshed_has_changed and self._recompile_tests(
                "Vendored typeshed has changed, recompiling tests..."
            ):
                self._run_mdtest(self.filters)

            for path in new_md_files | changed_md_files | rejected_snapshot_md_files:
                self._run_mdtests_for_file(path)


def main() -> None:
    parser = argparse.ArgumentParser(
        description="A runner for Markdown-based tests for ty"
    )
    parser.add_argument(
        "filters",
        nargs="*",
        help="Partial paths or mangled names, e.g., 'loops/for.md' or 'loops_for'",
    )
    parser.add_argument(
        "--enable-external",
        "-e",
        action="store_true",
        help="Enable tests with external dependencies",
    )
    parser.add_argument(
        "--no-lockfile-upgrades",
        action="store_true",
        help="By default, lockfiles will be upgraded when dependency requirements in the Markdown test change."
        + " Set this flag to never upgrade any lockfiles.",
    )

    args = parser.parse_args()

    try:
        runner = MDTestRunner(
            filters=args.filters,
            enable_external=args.enable_external,
            upgrade_lockfiles=not args.no_lockfile_upgrades,
        )
        runner.watch()
    except KeyboardInterrupt:
        print()


if __name__ == "__main__":
    main()



"""This module provides access to the Unix password database.
It is available on all Unix versions.

Password database entries are reported as 7-tuples containing the following
items from the password database (see `<pwd.h>'), in order:
pw_name, pw_passwd, pw_uid, pw_gid, pw_gecos, pw_dir, pw_shell.
The uid and gid items are integers, all others are strings. An
exception is raised if the entry asked for cannot be found.
"""

import sys
from _typeshed import structseq
from typing import Any, Final, final

if sys.platform != "win32":
    @final
    class struct_passwd(structseq[Any], tuple[str, str, int, int, str, str, str]):
        """pwd.struct_passwd: Results from getpw*() routines.

        This object may be accessed either as a tuple of
          (pw_name,pw_passwd,pw_uid,pw_gid,pw_gecos,pw_dir,pw_shell)
        or via the object attributes as named in the above tuple.
        """

        if sys.version_info >= (3, 10):
            __match_args__: Final = ("pw_name", "pw_passwd", "pw_uid", "pw_gid", "pw_gecos", "pw_dir", "pw_shell")

        @property
        def pw_name(self) -> str:
            """user name"""

        @property
        def pw_passwd(self) -> str:
            """password"""

        @property
        def pw_uid(self) -> int:
            """user id"""

        @property
        def pw_gid(self) -> int:
            """group id"""

        @property
        def pw_gecos(self) -> str:
            """real name"""

        @property
        def pw_dir(self) -> str:
            """home directory"""

        @property
        def pw_shell(self) -> str:
            """shell program"""

    def getpwall() -> list[struct_passwd]:
        """Return a list of all available password database entries, in arbitrary order.

        See help(pwd) for more on password database entries.
        """

    def getpwuid(uid: int, /) -> struct_passwd:
        """Return the password database entry for the given numeric user ID.

        See `help(pwd)` for more on password database entries.
        """

    def getpwnam(name: str, /) -> struct_passwd:
        """Return the password database entry for the given user name.

        See `help(pwd)` for more on password database entries.
        """



{x for i in ll}
{b for c in d if x in w if y and yy if z}
{a for b in c if d and e for f in j if k > h}
{a for b in c if d and e async for f in j if k > h}
{a for a, b in G}

# Non-parenthesized iter/if for the following expressions aren't allowed, so make sure
# it parses correctly for the parenthesized cases
{x for x in (yield y)}
{x for x in (yield from y)}
{x for x in (lambda y: y)}
{x for x in data if (yield y)}
{x for x in data if (yield from y)}
{x for x in data if (lambda y: y)}



"""Shared OS X support functions."""

from collections.abc import Iterable, Sequence
from typing import Final, TypeVar

_T = TypeVar("_T")
_K = TypeVar("_K")
_V = TypeVar("_V")

__all__ = ["compiler_fixup", "customize_config_vars", "customize_compiler", "get_platform_osx"]

_UNIVERSAL_CONFIG_VARS: Final[tuple[str, ...]]  # undocumented
_COMPILER_CONFIG_VARS: Final[tuple[str, ...]]  # undocumented
_INITPRE: Final[str]  # undocumented

def _find_executable(executable: str, path: str | None = None) -> str | None:  # undocumented
    """Tries to find 'executable' in the directories listed in 'path'.

    A string listing directories separated by 'os.pathsep'; defaults to
    os.environ['PATH'].  Returns the complete filename or None if not found.
    """

def _read_output(commandstring: str, capture_stderr: bool = False) -> str | None:  # undocumented
    """Output from successful command execution or None"""

def _find_build_tool(toolname: str) -> str:  # undocumented
    """Find a build tool on current path or using xcrun"""

_SYSTEM_VERSION: Final[str | None]  # undocumented

def _get_system_version() -> str:  # undocumented
    """Return the OS X system version as a string"""

def _remove_original_values(_config_vars: dict[str, str]) -> None:  # undocumented
    """Remove original unmodified values for testing"""

def _save_modified_value(_config_vars: dict[str, str], cv: str, newvalue: str) -> None:  # undocumented
    """Save modified and original unmodified value of configuration var"""

def _supports_universal_builds() -> bool:  # undocumented
    """Returns True if universal builds are supported on this system"""

def _find_appropriate_compiler(_config_vars: dict[str, str]) -> dict[str, str]:  # undocumented
    """Find appropriate C compiler for extension module builds"""

def _remove_universal_flags(_config_vars: dict[str, str]) -> dict[str, str]:  # undocumented
    """Remove all universal build arguments from config vars"""

def _remove_unsupported_archs(_config_vars: dict[str, str]) -> dict[str, str]:  # undocumented
    """Remove any unsupported archs from config vars"""

def _override_all_archs(_config_vars: dict[str, str]) -> dict[str, str]:  # undocumented
    """Allow override of all archs with ARCHFLAGS env var"""

def _check_for_unavailable_sdk(_config_vars: dict[str, str]) -> dict[str, str]:  # undocumented
    """Remove references to any SDKs not available"""

def compiler_fixup(compiler_so: Iterable[str], cc_args: Sequence[str]) -> list[str]:
    """
    This function will strip '-isysroot PATH' and '-arch ARCH' from the
    compile flags if the user has specified one them in extra_compile_flags.

    This is needed because '-arch ARCH' adds another architecture to the
    build, without a way to remove an architecture. Furthermore GCC will
    barf if multiple '-isysroot' arguments are present.
    """

def customize_config_vars(_config_vars: dict[str, str]) -> dict[str, str]:
    """Customize Python build configuration variables.

    Called internally from sysconfig with a mutable mapping
    containing name/value pairs parsed from the configured
    makefile used to build this interpreter.  Returns
    the mapping updated as needed to reflect the environment
    in which the interpreter is running; in the case of
    a Python from a binary installer, the installed
    environment may be very different from the build
    environment, i.e. different OS levels, different
    built tools, different available CPU architectures.

    This customization is performed whenever
    distutils.sysconfig.get_config_vars() is first
    called.  It may be used in environments where no
    compilers are present, i.e. when installing pure
    Python dists.  Customization of compiler paths
    and detection of unavailable archs is deferred
    until the first extension module build is
    requested (in distutils.sysconfig.customize_compiler).

    Currently called from distutils.sysconfig
    """

def customize_compiler(_config_vars: dict[str, str]) -> dict[str, str]:
    """Customize compiler path and configuration variables.

    This customization is performed when the first
    extension module build is requested
    in distutils.sysconfig.customize_compiler.
    """

def get_platform_osx(_config_vars: dict[str, str], osname: _T, release: _K, machine: _V) -> tuple[str | _T, str | _K, str | _V]:
    """Filter values for get_platform()"""



"""Python interfaces to XML parsers.

This package contains one module:

expat -- Python wrapper for James Clark's Expat parser, with namespace
         support.

"""

from xml.parsers import expat as expat



"""High-level support for working with threads in asyncio"""

from collections.abc import Callable
from typing import TypeVar
from typing_extensions import ParamSpec

# Keep asyncio.__all__ updated with any changes to __all__ here
__all__ = ("to_thread",)
_P = ParamSpec("_P")
_R = TypeVar("_R")

async def to_thread(func: Callable[_P, _R], /, *args: _P.args, **kwargs: _P.kwargs) -> _R:
    """Asynchronously run function *func* in a separate thread.

    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propagated,
    allowing context variables from the main thread to be accessed in the
    separate thread.

    Return a coroutine that can be awaited to get the eventual result of *func*.
    """



"""Filename matching with shell patterns.

fnmatch(FILENAME, PATTERN) matches according to the local convention.
fnmatchcase(FILENAME, PATTERN) always takes case in account.

The functions operate by translating the pattern into a regular
expression.  They cache the compiled regular expressions for speed.

The function translate(PATTERN) returns a regular expression
corresponding to PATTERN.  (It does not compile it.)
"""

import sys
from collections.abc import Iterable
from typing import AnyStr

__all__ = ["filter", "fnmatch", "fnmatchcase", "translate"]
if sys.version_info >= (3, 14):
    __all__ += ["filterfalse"]

def fnmatch(name: AnyStr, pat: AnyStr) -> bool:
    """Test whether FILENAME matches PATTERN.

    Patterns are Unix shell style:

    *       matches everything
    ?       matches any single character
    [seq]   matches any character in seq
    [!seq]  matches any char not in seq

    An initial period in FILENAME is not special.
    Both FILENAME and PATTERN are first case-normalized
    if the operating system requires it.
    If you don't want this, use fnmatchcase(FILENAME, PATTERN).
    """

def fnmatchcase(name: AnyStr, pat: AnyStr) -> bool:
    """Test whether FILENAME matches PATTERN, including case.

    This is a version of fnmatch() which doesn't case-normalize
    its arguments.
    """

def filter(names: Iterable[AnyStr], pat: AnyStr) -> list[AnyStr]:
    """Construct a list from those elements of the iterable NAMES that match PAT."""

def translate(pat: str) -> str:
    """Translate a shell PATTERN to a regular expression.

    There is no way to quote meta-characters.
    """

if sys.version_info >= (3, 14):
    def filterfalse(names: Iterable[AnyStr], pat: AnyStr) -> list[AnyStr]:
        """Construct a list from those elements of the iterable NAMES that do not match PAT."""



"""Convert a NT pathname to a file URL and vice versa.

This module only exists to provide OS-specific code
for urllib.requests, thus do not use directly.
"""

import sys
from typing_extensions import deprecated

if sys.version_info >= (3, 14):
    @deprecated("The `nturl2path` module is deprecated since Python 3.14.")
    def url2pathname(url: str) -> str:
        """OS-specific conversion from a relative URL of the 'file' scheme
        to a file system path; not recommended for general use.
        """

    @deprecated("The `nturl2path` module is deprecated since Python 3.14.")
    def pathname2url(p: str) -> str:
        """OS-specific conversion from a file system path to a relative URL
        of the 'file' scheme; not recommended for general use.
        """

else:
    def url2pathname(url: str) -> str:
        """OS-specific conversion from a relative URL of the 'file' scheme
        to a file system path; not recommended for general use.
        """

    def pathname2url(p: str) -> str:
        """OS-specific conversion from a file system path to a relative URL
        of the 'file' scheme; not recommended for general use.
        """
